#include <iostream>
#include "shapes.h"

//============================================================================

bool Plane::
quick_intersect(const Ray& ray, float tmin, float tmax) const
{
   float dn=dot(ray.direction, normal);
   if(dn==0) return false;
   float t=-dot(ray.origin-point,normal)/dn;
   return (t>=tmin && t<=tmax);
}

bool Plane::
intersect(const Ray& ray, float tmin, float tmax, float &t, Vec3f& surface_normal, Material& surface_material) const
{
   float dn=dot(ray.direction, normal);
   if(dn==0) return false; // if parallel, there's no intersection
   float t0=-dot(ray.origin-point,normal)/dn; // otherwise, the ray hits at this parameter value
   if(t0>=tmin && t0<=tmax){ // if it's in range, we hit!
      t=t0;
      // make sure we return a surface normal which faces the ray
      if(dot(normal,ray.direction)<0) surface_normal=normal;
      else surface_normal=-normal; // reverse the surface normal to face the ray
      surface_material=material;
      return true;
   }else
      return false;
}

BoundingBox Plane::
bounding_box(void) const
{
   // the max call is there to prevent a completely empty bounding box from being created
   // for a plane which is parallel to one of the axes. That could cause the primitive bounding box
   // intersection test to fail, which would be bad.
   Vec3f vector_bound(bound*max(std::sqrt((1-sqr(normal[0]))),1e-6f),
                      bound*max(std::sqrt((1-sqr(normal[1]))),1e-6f),
                      bound*max(std::sqrt((1-sqr(normal[2]))),1e-6f));
   return BoundingBox(point-vector_bound, point+vector_bound);
}

//============================================================================

bool Sphere::
quick_intersect(const Ray& ray, float tmin, float tmax) const
{
   // @@@ ADD CODE HERE
   // You need to determine if the given ray intersects the sphere
   // at a parameter value in the interval [tmin,tmax]
   // @@@

   return false; // this is not the correct code, but for now stops the compiler complaining
}

bool Sphere::
intersect(const Ray& ray, float tmin, float tmax, float &t, Vec3f& surface_normal, Material& surface_material) const
{
   // @@@ ADD CODE HERE
   // You need to determine if the given ray intersects the sphere
   // at a parameter value in the interval [tmin,tmax].
   // If there is an intersection, set t equal to the first intersection in the interval,
   // set surface_normal to the appropriate normal there (the one that points opposite to the ray)
   // and set surface_material to the sphere's material variable.
   // @@@

   return false; // this is not the correct code, but for now stops the compiler complaining
}

BoundingBox Sphere::
bounding_box(void) const
{
   // @@@ ADD CODE HERE
   // You need to return a bounding box which contains the sphere
   // @@@

   return BoundingBox(); // this is not the correct code, but for now stops the compiler complaining
}

//============================================================================

void TriangleMesh::
build_tree(void)
{
   std::vector<BoundingBox> box(tri.size());
   for(unsigned int f=0; f<tri.size(); ++f){
      int i, j, k; assign(tri[f], i, j, k);
      box[f].build_from_points(x[i], x[j], x[k]);
      box[f].enlarge(epsilon); // to avoid rounding problems later in intersection testing
   }
   tree.construct_from_leaf_boxes(tri.size(), &box[0]);
}

bool TriangleMesh::
quick_intersect(const Ray& ray, float tmin, float tmax) const
{
   // first check if the ray hits the root's bounding box
   float t;
   if(!tree.box.intersect(ray, tmin, tmax, t))
      return false;
   // if we get here it does hit; do a depth first search of the tree
   std::vector<const BoundingBoxTree*> stack;
   stack.push_back(&tree);
   float alpha, beta, gamma;
   while(!stack.empty()){
      const BoundingBoxTree *node=stack.back();
      stack.pop_back();
      // check if there's any primitive geometry to check in this node
      for(unsigned int i=0; i<node->index.size(); ++i){
         if(intersect_triangle(node->index[i], ray, tmin, tmax, t, alpha, beta, gamma)) return true;
      }
      // check if there are children nodes (with boxes intersecting the ray) to search later
      for(unsigned int i=0; i<node->children.size(); ++i){
         if(node->children[i]->box.intersect(ray, tmin, tmax, t))
            stack.push_back(node->children[i]); 
      }
   }
   return false; // got through search of tree without finding an intersection
}

bool TriangleMesh::
intersect(const Ray& ray, float tmin, float tmax, float &t, Vec3f& surface_normal, Material& surface_material) const
{
   // first check if the ray hits the root's bounding box
   float tbox;
   if(!tree.box.intersect(ray, tmin, tmax, tbox)) return false;
   // if we get here it does hit; do a depth first search of the tree for the first intersection if there is one
   bool found_intersection=false;
   std::vector<const BoundingBoxTree*> stack;
   stack.push_back(&tree);
   float t0, alpha, beta, gamma;
   while(!stack.empty()){
      const BoundingBoxTree *node=stack.back();
      stack.pop_back();
      // check if there's any primitive geometry to check in this node
      for(unsigned int i=0; i<node->index.size(); ++i){
         int f=node->index[i];
         if(intersect_triangle(f, ray, tmin, tmax, t0, alpha, beta, gamma)){
            t=t0;
            tmax=t0; // eliminate any further intersections from consideration
            if(normal.empty()){
               surface_normal=normalized(cross(x[tri[f][1]]-x[tri[f][0]], x[tri[f][2]]-x[tri[f][0]]));
            }else{
               surface_normal=normalized(alpha*normal[tri[f][0]]+beta*normal[tri[f][1]]+gamma*normal[tri[f][2]]);
            }
            found_intersection=true;
         }
      }
      // check if there are children nodes (with boxes intersecting the ray) to search later
      // Note: we could optimize this to put the children in reverse order of tbox on the stack,
      // but for simplicity's sake we won't do the sort.
      for(unsigned int i=0; i<node->children.size(); ++i){
         if(node->children[i]->box.intersect(ray, tmin, tmax, tbox)){
            stack.push_back(node->children[i]); 
         }
      }
   }
   if(found_intersection){
      if(dot(ray.direction, surface_normal)>0) surface_normal=-surface_normal;
      surface_material=material;
      return true;
   }else{
      return false;
   }
}

BoundingBox TriangleMesh::
bounding_box(void) const
{
   assert(!tree.index.empty() || !tree.children.empty()); // check that tree was built before calling this
   return tree.box;
}

// This is the core routine for TriangleMesh: how to intersect a ray with a single triangle.
bool TriangleMesh::
intersect_triangle(unsigned int f, const Ray& ray, float tmin, float tmax, float& t, float& alpha, float& beta, float& gamma) const
{
   // get the vertex indices (i, j, k) from tri[f]: the triangle has vertices x[i], x[j], and x[k]
   int i, j, k; assign(tri[f], i, j, k);
   // @@@ ADD CODE HERE
   // You need to check if the ray intersects this triangle at a parameter value in the range [tmin,tmax].
   // If it does, set t equal to the parameter value of the intersection, and also set
   // alpha, beta and gamma to the barycentric coordinates of the intersection point.
   // Note: include a provision for rounding error, saying the ray hits the triangle when all barycentric
   // coordinates are greater than or equal to -epsilon, not just zero.
   // @@@

   return false; // this is not the correct code, but for now stops the compiler complaining
}

//============================================================================

void AggregateShape::
build_tree(void)
{
   std::vector<BoundingBox> box(subshape.size());
   for(unsigned int i=0; i<subshape.size(); ++i){
      box[i]=subshape[i]->bounding_box();
   }
   tree.construct_from_leaf_boxes(subshape.size(), &box[0]);
}

bool AggregateShape::
quick_intersect(const Ray& ray, float tmin, float tmax) const
{
   // first check if the ray hits the root's bounding box
   float t;
   if(!tree.box.intersect(ray, tmin, tmax, t))
      return false;
   // if we get here it does hit; do a depth first search of the tree
   std::vector<const BoundingBoxTree*> stack;
   stack.push_back(&tree);
   while(!stack.empty()){
      const BoundingBoxTree *node=stack.back();
      stack.pop_back();
      // check if there's any primitive geometry to check in this node
      for(unsigned int i=0; i<node->index.size(); ++i){
         if(subshape[node->index[i]]->quick_intersect(ray, tmin, tmax)) return true;
      }
      // check if there are children nodes (with boxes intersecting the ray) to search later
      for(unsigned int i=0; i<node->children.size(); ++i){
         if(node->children[i]->box.intersect(ray, tmin, tmax, t))
            stack.push_back(node->children[i]); 
      }
   }
   return false; // got through search of tree without finding an intersection
}

bool AggregateShape::
intersect(const Ray& ray, float tmin, float tmax, float &t, Vec3f& surface_normal, Material& surface_material) const
{
   // first check if the ray hits the root's bounding box
   float t_test;
   if(!tree.box.intersect(ray, tmin, tmax, t_test)) return false;
   // if we get here it does hit; do a depth first search of the tree for the first intersection if there is one
   bool found_intersection=false;
   std::vector<const BoundingBoxTree*> stack;
   stack.push_back(&tree);
   while(!stack.empty()){
      const BoundingBoxTree *node=stack.back();
      stack.pop_back();
      // check if there's any primitive geometry to check in this node
      for(unsigned int i=0; i<node->index.size(); ++i){
         if(subshape[node->index[i]]->intersect(ray, tmin, tmax, t_test, surface_normal, surface_material)){
            assert(t_test==t_test); // guard against NaN
            assert(surface_normal==surface_normal); // guard against NaN
            t=t_test;
            tmax=t_test; // eliminate any further intersections from consideration
            found_intersection=true;
         }
      }
      // check if there are children nodes (with boxes intersecting the ray) to search later
      // Note: we could optimize this to put the children in reverse order of t_test on the stack
      for(unsigned int i=0; i<node->children.size(); ++i){
         if(node->children[i]->box.intersect(ray, tmin, tmax, t_test)){
            stack.push_back(node->children[i]); 
         }
      }
   }
   return found_intersection;
}

BoundingBox AggregateShape::
bounding_box(void) const
{
   assert(!tree.index.empty() || !tree.children.empty()); // check that tree was built before calling this
   return tree.box;
}

